/*---------------------------------------------------------------------------*
 *                                   IT++			             *
 *---------------------------------------------------------------------------*
 * Copyright (c) 1995-2003 by Tony Ottosson, Thomas Eriksson, Pl Frenger,   *
 * Tobias Ringstrm, and Jonas Samuelsson.                                   *
 *                                                                           *
 * Permission to use, copy, modify, and distribute this software and its     *
 * documentation under the terms of the GNU General Public License is hereby *
 * granted. No representations are made about the suitability of this        *
 * software for any purpose. It is provided "as is" without expressed or     *
 * implied warranty. See the GNU General Public License for more details.    *
 *---------------------------------------------------------------------------*/

/*!
  \file
  \brief Definitions of Filter classes and functions
  \author Hkan Eriksson and Thomas Eriksson

  1.15

  2003/05/22 08:55:16
*/

#ifndef __filter_h
#define __filter_h

#include <complex>
#include <cstdlib>
#include <iostream>
#include <algorithm>

#include "itconfig.h"
#include "base/vec.h"
#include "base/matfunc.h"
#include "base/specmat.h"
#include "base/itassert.h"

namespace itpp {

  /*!
    \defgroup filters Filter Classes and Functions
  */

  /*!
    \brief Virtual Filter Base Class.
    \ingroup filters

    The class is templated as follows:
    <ul>
    <li> \c T1 is the type of the input samples</li>
    <li> \c T2 is the type of the filter coefficients</li>
    <li> \c T3 is the type of the output samples</li>
    </ul>
  */
  template <class T1, class T2, class T3>
    class Filter {
    public:
    //! Default constructor
    Filter();
    //! Filter a single sample.
    virtual T3 operator()(T1 Sample) { return filter(Sample); }
    //! Filter a vector.
    virtual Vec<T3> operator()(const Vec<T1> &v);
    protected:
    /*!
      \brief Pure virtual filter function. This is where the real filtering is done. Implement this function to create a new filter.
    */
    virtual T3 filter(T1 Sample)=0;
  };

  /*!
    \brief Moving Average Filter Base Class.
    \ingroup filters

    This class implements a moving average (MA) filter
    according to
    \f[
    y(n) = coeff(0)*x(n) + coeff(1)*x(n-1) + ... + coeff(N)*x(n-N)
    \f]
    where \a coeff is the filter coefficients, \a x is the input and \a y is the output.

    When filtering a vector, the length of the output vector equals the length of the input vector.
    Internal states are kept in a filter memory. The first time the filter is used the internal states
    have been set to zero.

    The class is templated as follows:
    <ul>
    <li> \c T1 is the type of the input samples</li>
    <li> \c T2 is the type of the filter coefficients</li>
    <li> \c T3 is the type of the output samples</li>
    </ul>
  */
  template <class T1, class T2, class T3>
    class MA_Filter : public Filter<T1,T2,T3> {
    public:
    //! Class default constructor
    explicit MA_Filter();
    //! Class constructor setting the coefficients in the filter
    explicit MA_Filter(const Vec<T2> &Par);
    //! Class destructor
    virtual ~MA_Filter() { }
    //! Filter coefficient access function
    Vec<T2> get_coeffs() const { return coeffs; }
    //! Set the filter coefficients
    void set_coeffs(const Vec<T2> &c);
    //! Clears the filter memory
    void clear() { mem.clear(); }

    //protected:
    private:
    virtual T3 filter(T1 Sample);

    Vec<T3> mem;
    Vec<T2> coeffs;
    long inptr;
  };

  /*!
    \brief Autoregressive (AR) Filter Base Class.
    \ingroup filters

    This class implements a moving average (MA) filter
    according to
    \f[
    coeff(0)*y(n) = x(n) - coeff(1)*y(n-1) - ... - coeff(N)*y(n-N)
    \f]
    where \a coeff is the filter coefficients, \a x is the input and \a y is the output.

    When filtering a vector, the length of the output vector equals the length of the input vector.
    Internal states are kept in a filter memory. The first time the filter is used the internal states
    have been set to zero.

    The class is templated as follows:
    <ul>
    <li> \c T1 is the type of the input samples</li>
    <li> \c T2 is the type of the filter coefficients</li>
    <li> \c T3 is the type of the output samples</li>
    </ul>
  */
  template <class T1, class T2, class T3>
    class AR_Filter : public Filter<T1,T2,T3> {
    public:
    //! Class constructor
    explicit AR_Filter();
    //! Class constructor setting the coefficnebts in the filter
    explicit AR_Filter(const Vec<T2> &Par);
    //! Class destructor
    virtual ~AR_Filter() { }
    //! Filter coefficient access function
    Vec<T2> get_coeffs() const { return coeffs; }
    //! Set the filter coefficients (and order)
    void set_coeffs(const Vec<T2> &c);
    //! Clears the filter memory
    void clear() { mem.clear(); }

    //protected:
    private:
    virtual T3 filter(T1 Sample);

    Vec<T3> mem;
    Vec<T2> coeffs;
    long inptr;
  };

  /*!
    \brief ARMA filter (Auto-Regressive Moving Average)
    \ingroup filters

    This function is templated as follows: \c T1 is the type of the input samples,
    \c T2 is the type of the filter coefficients, and \c T3 is the type of the output samples.
  */
  template <class T1, class T2, class T3>
    Vec<T3> filter(const Vec<T2> &b, const Vec<T2> &a, const Vec<T1> &input);

  /*!
    \brief FIR MA filter (Finite Impulse Response Moving Average)
    \ingroup filters

    This function is templated as follows: \c T1 is the type of the input samples,
    \c T2 is the type of the filter coefficients, and \c T3 is the type of the output samples.
  */
  template <class T1, class T2, class T3>
    Vec<T3> filter(const Vec<T2> &b, int one, const Vec<T1> &input);

  /*!
    \brief IIR AR filter (Infinite Impulse Response Auto-Regressive)
    \ingroup filters

    This function is templated as follows: \c T1 is the type of the input samples,
    \c T2 is the type of the filter coefficients, and \c T3 is the type of the output samples.
  */
  template <class T1, class T2, class T3>
    Vec<T3> filter(int one, const Vec<T2> &a, const Vec<T1> &input);

  /*!
    \brief Design a Nth order FIR filter with cut-off frequency \c cutoff using the window method.
    \ingroup filters
  */
  vec fir1(long N, double cutoff);

  //-------------------------------------------------------------------------------------------------
  // Implementation of templated functions starts here
  //-------------------------------------------------------------------------------------------------

  //---------------------- class Filter ----------------------------

  template <class T1, class T2, class T3>
    Filter<T1,T2,T3>::Filter()
    {
    }

  template <class T1, class T2, class T3>
    Vec<T3> Filter<T1,T2,T3>::operator()(const Vec<T1> &x)
    {
      Vec<T3> y(x.length());

      for (long i=0;i<x.length();i++) {
	y[i]=filter(x[i]);
      }
      return y;
    }

  //-------------------------- class MA_Filter ---------------------------------

  template <class T1, class T2,class T3>
    MA_Filter<T1,T2,T3>::MA_Filter() : Filter<T1,T2,T3>()
    {
      inptr=0;
    }

  template <class T1, class T2,class T3>
    MA_Filter<T1,T2,T3>::MA_Filter(const Vec<T2> &Par) : Filter<T1,T2,T3>()
    {
      coeffs=Par;
      inptr=0;
      mem.set_length(coeffs.length());
      mem.clear();
    }

  template <class T1, class T2, class T3>
    T3 MA_Filter<T1,T2,T3>::filter(T1 Sample)
    {
      T3  s=0;
      long i,L;

      mem._elem(inptr)=Sample;
      L=mem.length()-inptr;

      for (i=0;i<L;i++) {
	s+=coeffs._elem(i)*mem._elem(inptr+i);
      }
      for (i=0;i<inptr;i++) {
	s+=coeffs._elem(L+i)*mem._elem(i);
      }
      inptr=inptr-1;if (inptr<0) inptr+=mem.length();
      return s;
    }

  template <class T1, class T2, class T3>
    void MA_Filter<T1,T2,T3>::set_coeffs(const Vec<T2> &Par)
    {
      coeffs=Par;
      if (mem.length()!=coeffs.length()) {
	inptr=0;
	mem.set_length(coeffs.length(),true);
	mem.clear();
      }
    }

  //---------------------- class AR_Filter ----------------------------------

  template <class T1, class T2, class T3>
    AR_Filter<T1,T2,T3>::AR_Filter() : Filter<T1,T2,T3>()
    {
      inptr=0;
    }

  template <class T1, class T2, class T3>
    AR_Filter<T1,T2,T3>::AR_Filter(const Vec<T2> &Par) : Filter<T1,T2,T3>()
    {
      coeffs=Par.right(length(Par)-1)/Par(0);
      inptr=0;
      mem.set_length(coeffs.length(),false);
      mem.clear();
    }

  template <class T1, class T2, class T3>
    void AR_Filter<T1,T2,T3>::set_coeffs(const Vec<T2> &Par)
    {
      coeffs=Par.right(length(Par)-1)/Par(0);
      if (mem.length()!=coeffs.length()) {
	inptr=0;
	mem.set_length(coeffs.length(),false);
	mem.clear();
      }
    }

  template <class T1, class T2, class T3>
    T3 AR_Filter<T1,T2,T3>::filter(T1 Sample)
    {
      T3  s=0;
      long i,L;

      L=length(mem)-inptr;
      for (i=0;i<L;i++) {
	s+=coeffs._elem(i)*mem._elem(inptr+i);
      }
      for (i=0;i<inptr;i++) {
	s+=coeffs._elem(L+i)*mem._elem(i);
      }
      inptr=inptr-1;if (inptr<0) inptr+=length(mem);

      s=Sample-s;
      mem._elem(inptr)=s;

      return s;
    }

  //------------------------------ filter functions -----------------------------

  template <class T1, class T2, class T3>
    Vec<T3> filter(const Vec<T2> &b, const Vec<T2> &a, const Vec<T1> &input)
    {
      Vec<T2> atemp=a, btemp=b;
      int inputlength = input.length();
      int blength = b.length();
      int alength = a.length();
      int memorylength = (alength > blength ? (alength-1) : (blength-1));
      Vec<T3> chanmemory(memorylength);
      Vec<T3> output(inputlength);
      T3 memoryinput, temp;
      T2 norm = a(0);
      int i, j;

      chanmemory.clear();
      atemp /= norm;
      btemp /= norm;
      for (i = 0; i < inputlength; i++) {
	temp = 0;
	for (j=1; j<alength; j++)
	  temp += chanmemory(j-1) * atemp(j);
	memoryinput = input(i) - temp;
	temp = memoryinput * btemp(0);
	for (j=1; j<blength; j++)
	  temp += chanmemory(j-1) * btemp(j);
	output(i) = temp;
	chanmemory.shift_right(memoryinput);
      }
      return( output );
    }

  template <class T1, class T2, class T3>
    Vec<T3> filter(const Vec<T2> &b, int one, const Vec<T1> &input)
    {
      it_error_if(one!=1,"filter error");
      Vec<T2> btemp=b;
      int inputlength = input.length();
      int blength = b.length();
      int alength = 1;
      int memorylength = (alength > blength ? (alength-1) : (blength-1));
      Vec<T3> chanmemory(memorylength);
      Vec<T3> output(inputlength);
      T3 memoryinput, temp;
      int i, j;

      chanmemory.clear();
      for (i = 0; i < inputlength; i++) {
	temp = 0;
	memoryinput = input(i) - temp;
	temp = memoryinput * btemp(0);
	for (j=1; j<blength; j++)
	  temp += chanmemory(j-1) * btemp(j);
	output(i) = temp;
	chanmemory.shift_right(memoryinput);
      }
      return( output );
    }

  template <class T1, class T2, class T3>
    Vec<T3> filter(int one, const Vec<T2> &a, const Vec<T1> &input)
    {
      it_error_if(one!=1,"filter error");
      Vec<T2> atemp=a;
      T2 btemp=1;
      int inputlength = input.length();
      int blength = 1;
      int alength = a.length();
      int memorylength = (alength > blength ? (alength-1) : (blength-1));
      Vec<T3> chanmemory(memorylength);
      Vec<T3> output(inputlength);
      T3 memoryinput, temp;
      T2 norm = a(0);
      int i, j;

      chanmemory.clear();
      atemp /= norm;
      btemp /= norm;
      for (i = 0; i < inputlength; i++) {
	temp = 0;
	for (j=1; j<alength; j++)
	  temp += chanmemory(j-1) * atemp(j);
	memoryinput = input(i) - temp;
	temp = memoryinput * btemp;
	output(i) = temp;
	chanmemory.shift_right(memoryinput);
      }
      return( output );
    }

} //namespace itpp

#endif // __filter_h
