/*---------------------------------------------------------------------------*
 *                                   IT++			             *
 *---------------------------------------------------------------------------*
 * Copyright (c) 1995-2004 by Tony Ottosson, Thomas Eriksson, Pl Frenger,   *
 * Tobias Ringstrm, and Jonas Samuelsson.                                   *
 *                                                                           *
 * Permission to use, copy, modify, and distribute this software and its     *
 * documentation under the terms of the GNU General Public License is hereby *
 * granted. No representations are made about the suitability of this        *
 * software for any purpose. It is provided "as is" without expressed or     *
 * implied warranty. See the GNU General Public License for more details.    *
 *---------------------------------------------------------------------------*/

/*! 
  \file 
  \brief Definitions of Fourier, Hadamard, Walsh-Hadamard, and 2D Hadamard transforms
  \author Tony Ottosson

  This file uses the FFTW package version 3.0.1 that is distributed under the GNU GPL licence. 
  For details see http://www.fftw.org.

  1.14

  2004/09/28 11:48:19
*/

#ifndef __transforms_h
#define __transforms_h

#include "base/vec.h"
#include "base/mat.h"

namespace itpp {

  /*! \defgroup fft Fast Fourier Transforms (FFT)
    \brief One dimensional fast fourier transform
    \author Tony Ottosson

    The functions \code X = fft(x) \endcode and \code x = ifft(X) \endcode are
    the fourier and inverse fourier transforms of size \a N defined as:
    \f[
    X(k) = \sum_{j=0}^{N-1} x(j) e^{-2\pi j k \cdot i / N}
    \f]

    \f[
    x(j) = \frac{1}{N} \sum_{k=0}^{N-1} X(k) e^{2\pi j k \cdot i / N}
    \f]

    \code Y = fft(X, N) \endcode performs zero-padding up to size N and then
    performs an N-size fft.

    The implementation is built upon the FFTW library (www.fftw.org) which is
    a big set of fast and reliable fft routines used e.g. in matlab.

    The routine is fastest for powers of two. Furthermore, the second time you
    call the routine with the same size, the calculation is much faster due to
    many things were calculated and stored the first time the routine was called.
  */

#ifndef NO_FFTW

  //!\addtogroup fft
  //!@{

  //! Fast Fourier Transform
  void fft(const cvec &in, cvec &out);
  //! Fast Fourier Transform
  cvec fft(const cvec &in);
  //! Fast Fourier Transform, with zero-padding up to size N
  cvec fft(const cvec &in, int N);
  //! Inverse Fast Fourier Transform
  void ifft(const cvec &in, cvec &out);
  //! Inverse Fast Fourier Transform
  cvec ifft(const cvec &in);
  //! Inverse Fast Fourier Transform, with zero-padding up to size N
  cvec ifft(const cvec &in, int N);

  //! Real Fast Fourier Transform
  void fft_real(const vec& in, cvec &out);
  //! Real Fast Fourier Transform
  cvec fft_real(const vec& in);
  //! Inverse Fast Fourier Transform. Assumes even size.
  void ifft_real(const cvec &in, vec &out);
  //! Inverse Fast Fourier Transform. Assumes even size.
  vec ifft_real(const cvec &in);
  //!@}

#endif // NO_FFTW

  /*! \defgroup fht Fast Hadamard Transforms */

  //!\addtogroup fht
  //!@{

  //! Fast Hadamard Transform
  template <class T> Vec<T> dht(const Vec<T> &v);
  //! Fast Hadamard Transform
  template <class T> void dht(const Vec<T> &vin, Vec<T> &vout);
  //! Fast Hadamard Transform - memory efficient. Stores the result in \c v
  template <class T> void self_dht(Vec<T> &v);

  //! Fast Walsh Hadamard Transform
  template <class T> Vec<T> dwht(const Vec<T> &v);
  //! Fast Walsh Hadamard Transform
  template <class T> void dwht(const Vec<T> &vin, Vec<T> &vout);
  //! Fast Walsh Hadamard Transform - memory efficient. Stores the result in \c v
  template <class T> void self_dwht(Vec<T> &v);

  //! Fast 2D Hadamard Transform
  template <class T> Mat<T> dht2(const Mat<T> &m);
  //! Fast 2D Walsh Hadamard Transform
  template <class T> Mat<T> dwht2(const Mat<T> &m);
  //!@}

} //namespace itpp

#endif // __transforms_h
